/***************************************************************************\
  "inputmanager.hpp" - Input Manager for Buffered Input (Phase A)
  Revision 9; 10 Nov 2002 - 15 March 2004
  Copyright (C) 2002-2003  Kevin P. Albrecht

  References:
    MCD 2.1.1; 2.1.1.1; 2.1.12.1 (Input Buffer Stack)
\***************************************************************************/

#ifndef INPUTMANAGER_HPP
#define INPUTMANAGER_HPP

/////////////////////////
// DEFINES //////////////
/////////////////////////

#define INPUTMANAGER_EOF             0x7F
#define INPUTMANAGER_DIR_DELIM       '\\'

/////////////////////////
// INCLUDES /////////////
/////////////////////////

// Standard //

// Local //

/////////////////////////
// DECLARATIONS /////////
/////////////////////////

class Buffer
{
  public:
    Buffer ();
    ~Buffer ();

    FILE *file; // input file
    char *filename; // filename
    char text[GLOBAL_INPUTMANAGER_LINE_LEN]; // text buffer
    int pos; // current position in mText
    char curr; // current character
    int line; // current line number in file

    Buffer *next; // pointer to the next buffer in the stack
};

class InputManager
{
  public:
    InputManager ();
    ~InputManager ();

    void include (char *filename);
    char get ();
    void seek ();

    // THIS SHOULD BE PRIVATE, BUT IT IS NEEDED IN "ICM.HPP"
    //   FOR LINE PRINTING
    Buffer *mBufferStackTop;

  private:
    char popLine ();

    char *mDirectories[GLOBAL_INPUTMANAGER_MAX_DIRECTORIES]; // array of all the directories included
    int miDirectories;
};

/////////////////////////
// DEFINITIONS //////////
/////////////////////////

Buffer::Buffer ()
{
  // empty
}

//// Free the buffer
Buffer::~Buffer ()
{
  k_del_s (filename);
  fclose (file);
}

//// Constructor
InputManager::InputManager ()
{
  mBufferStackTop = NULL;

  for (int i = 0; i < GLOBAL_INPUTMANAGER_MAX_DIRECTORIES; i++)
  {
    mDirectories[i] = NULL;
  }

  miDirectories = 1;
}

//// Deconstructor
InputManager::~InputManager ()
{
  Buffer *temp;

  // free the buffer stack //
  while (mBufferStackTop)
  {
    temp = mBufferStackTop->next;
    delete mBufferStackTop;
    mBufferStackTop = temp;
  }

  for (; miDirectories > 0; miDirectories--)
  {
    k_del_s (mDirectories[miDirectories]);
  }
}

//// Include a new file
void InputManager::include (char *aFilename)
{
  Buffer *tempnew;
  char *sTemp;
  int i, iLast;

  // if there is already a file open, set the line number //
  if (mBufferStackTop)
  {
    mBufferStackTop->line = Global::nCurrentLine;
  }

  // reset the current line and file information //
  Global::nCurrentLine = 1;
  k_copy (Global::sCurrentFilename, aFilename);

  // Create new buffer //
  tempnew = new Buffer ();

  // Find last directory delimiter in filename //
  iLast = -1;
  i = k_posc (aFilename, INPUTMANAGER_DIR_DELIM, 0);
  while (i != -1)
  {
    iLast = i;
    i = k_posc (aFilename, INPUTMANAGER_DIR_DELIM, i + 1);
  }

  // There is a directory delimiter in the filename //
  if (iLast != -1)
  {
    // Add the included file's directory to the list //
    mDirectories[miDirectories++] = k_sub (aFilename, 0, iLast + 1);
  }

  // Assign values to new buffer //
  tempnew->filename = NULL;
  for (i = 0; i < miDirectories; i++)
  {
    if (tempnew->filename != NULL) k_del_s (tempnew->filename);
    if (mDirectories[i] != NULL)
    {
      tempnew->filename = k_new_s (strlen (aFilename) + strlen (mDirectories[i]));
      k_copy (tempnew->filename, mDirectories[i]);
      k_cat (tempnew->filename, aFilename);
    }
    else
    {
      tempnew->filename = k_new_s (strlen (aFilename));
      k_copy (tempnew->filename, aFilename);
    }
    tempnew->file = fopen (tempnew->filename, "r");

    if (tempnew->file) break;
    else
    {
      // Allow .JUL files to be specified without using "*.JUL" //
      sTemp = tempnew->filename;
      tempnew->filename = k_new_s (strlen (sTemp) + 4);
      k_copy (tempnew->filename, sTemp);
      k_cat (tempnew->filename, ".jul");
      tempnew->file = fopen (tempnew->filename, "r");
      k_del_s (sTemp);

      if (tempnew->file) break;
    }
  }

  if (!tempnew->file)
  {
    err (false, EP_FILE_OPEN_FAILED, aFilename);
    errm ("Also attempted with \".jul\"", NULL);
    errterm();
  }

  k_copy (tempnew->text, "");
  tempnew->curr = '\0';
  tempnew->pos = 0;
  
  // put the new filename in the global filename string //
  k_copy (Global::sCurrentFilename, tempnew->filename);

  // Put the new buffer on top of the stack //
  tempnew->next = mBufferStackTop;
  mBufferStackTop = tempnew;

  tempnew->curr = popLine ();
}

//// Return the current character
char InputManager::get ()
{
  if (!mBufferStackTop) return 0;

  return mBufferStackTop->curr;
}

//// Set the current character to the next character
void InputManager::seek ()
{
  if (!mBufferStackTop) return;

  if (mBufferStackTop->curr == INPUTMANAGER_EOF) // currently at end of file
  {
    return;
  }
  else if (mBufferStackTop->curr == 0) // currently at end of line
  {
    mBufferStackTop->curr = popLine ();
  }
  else
  {
    mBufferStackTop->curr = mBufferStackTop->text[++mBufferStackTop->pos];
  }
}

//// Put the next line from the file into the buffer.
//// Then return the line's first character.
char InputManager::popLine ()
{
  char *sTemp; // test to see if 'fgets' was successful
  Buffer *bufferTemp;

  if (!mBufferStackTop) return '\0';

  mBufferStackTop->pos = 0;

  if (feof (mBufferStackTop->file))
  {
    // This is only the end of the currently included file //
    if (mBufferStackTop->next)
    {
      bufferTemp = mBufferStackTop;
      mBufferStackTop = mBufferStackTop->next;
      delete bufferTemp;

      // set the current line and file inforamtion //
      Global::nCurrentLine = mBufferStackTop->line - 1;
      k_copy (Global::sCurrentFilename, mBufferStackTop->filename);

      return mBufferStackTop->text[mBufferStackTop->pos];
    }
    // This is the end the entire program //
    else
    {
      return INPUTMANAGER_EOF;
    }
  }
  else
  {
    sTemp = fgets (mBufferStackTop->text, GLOBAL_INPUTMANAGER_LINE_LEN, mBufferStackTop->file);

    //BUGFIX (14 July 2002 -- 0061):
    //  If the file was not terminated by a newline, the last char was erased
    if (mBufferStackTop->text[strlen(mBufferStackTop->text) - 1] < '\n')
    {
      mBufferStackTop->text[strlen(mBufferStackTop->text) - 1] = '\0';
    }
    //gnCurrentLine++;

    if (mBufferStackTop->text && sTemp) return mBufferStackTop->text[mBufferStackTop->pos];
	  else return 0;
  }
}

#endif
