/***************************************************************************\
  "julc.cpp" - Juliet Interpreter Controller Program
  (see below for version and date)
  Copyright (C) 2002-2004  Kevin P. Albrecht
\***************************************************************************/

/////////////////////////
// DEFINES //////////////
/////////////////////////

// Environment //
//#define ENV_GCC // GNU Compiler Collection C++
//#define ENV_LCC // LCC Win-32
//#define ENV_MSVC // Microsoft Visual C++
#define ENV_DMC  // Digital Mars C++

// Debug Status //
//#define GLOBAL_DEBUG
#define GLOBAL_RELEASE
//#define GLOBAL_DEBUG_VERBOSE

// Miscellaneous //
#define GLOBAL_MANTISSA_SIZE                13
#define GLOBAL_TOKEN_SIZE                   256
#define GLOBAL_MAX_SCOPES                   256
#define GLOBAL_MAX_SYMBOLS                  16384 // 2^14
#define GLOBAL_INPUTMANAGER_MAX_DIRECTORIES 16
#define GLOBAL_INPUTMANAGER_LINE_LEN        256
#define GLOBAL_LEXER_TAB_WIDTH              5
#define GLOBAL_TABLE_MAX_SYMBOLS            256
#define GLOBAL_ROUTINE_MAX_PARAMETERS       32
#define GLOBAL_ICM_MAX_INSTRUCTIONS         16384 // was 2^10, now 2^14
#define GLOBAL_TEMPGEN_MAX_TEMPS            256
#define GLOBAL_TEMPGEN_TEMP_NAME_LEN        64
#define GLOBAL_RUNTIME_ENGINE_CIN_LINE_LEN  256

/////////////////////////
// INCLUDES /////////////
/////////////////////////

// Standard (Standard C++ includes for program must go here) //
#include <iostream>
#include <stack>
using namespace std;

// LeakTracker (must go before any local code & after Std C++) //
#include "leaktracker.hpp"

// Local //
#include "k_string.hpp"
#include "global.hpp"
#include "error.hpp"
#include "symbol.hpp"

#include "unique.hpp"
UniqueGenerator *gUnique;

#include "scopemanager.hpp"
ScopeManager *gScopeManager;
Routine *gCurrRout;

#include "inputmanager.hpp"
InputManager *gInputManager;

#include "lexer.hpp"
Lexer *gLexer;

#include "statement.hpp"
#include "icm.hpp"
IntermediateCodeManager *gICM;

#include "parser.hpp"
Parser *gParser;

#include "icoutput.hpp"

/////////////////////////
// DECLARATIONS /////////
/////////////////////////

/////////////////////////
// DEFINITIONS //////////
/////////////////////////

int main (int nArgs, char *vsArgs[])
{
  // Leak Tracking //
  #ifdef GLOBAL_DEBUG
    leaktracker_init ();
  #endif

  char *sInputFilename = k_new_s (256); k_copy (sInputFilename, "I AM EMPTY!!");
  char *sOutputFilename = k_new_s (256); k_copy (sOutputFilename, "I AM EMPTY!!");

  // Initialize Global //
  Global::initialize();

  // Command-line Argument Handling //
  if (nArgs > 1)
  {
    for (int inArg = 1; inArg < nArgs; inArg++)
    {
      // '-a' //
      if (k_same (vsArgs[inArg], "-a"))
      {
        cout << "\n"
             << "========================= Juliet Compiler (JULC) =========================\n"
             << "\n"
             << "Notice:  Copyright (C) 2002-2004, Kevin P. Albrecht\n"
             << "E-mail:  kevin@albrecht.net\n"
             << "Website: http://kevin.alteu.com/\n"
             << "\n"
             << "JULC Version: 1.00.0401 (4 April 2004)\n"
             << "Juliet Spec:  2004.0 (4 April 2004)\n"
             << "Portia Spec:  2004.0 (4 April 2004)\n";
        return 0;
      }
      // '-l' //
      if (k_same (vsArgs[inArg], "-l"))
      {
        cout << "\n"
             << "Copyright (C) 2002-2004, Kevin P. Albrecht\n"
             << "\n"
             << "Permission is hereby granted by Kevin P. Albrecht (the \"Author\"), free of\n"
             << "charge, to any person obtaining a copy of this software and associated\n"
             << "documentation files (the \"Software\"), to deal in the Software without\n"
             << "restriction, including without limitation the rights to use, copy, modify,\n"
             << "merge, publish, distribute, sublicense, and/or sell copies of the\n"
             << "Software, and to permit persons to whom the Software is furnished to do\n"
             << "so, subject to the following conditions:\n"
             << "The above copyright notice and this permission notice shall be included\n"
             << "in all copies or substantial portions of the Software.\n"
             << "\n"
             << "THE SOFTWARE IS PROVIDED \"AS IS\", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR\n"
             << "IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,\n"
             << "FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL\n"
             << "THE AUTHOR BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN\n"
             << "AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN\n"
             << "CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.\n"
             << "\n"
             << "Except as contained in this notice, the name of the Author shall not be\n"
             << "used in advertising or otherwise to promote the sale, use or other\n"
             << "dealings in this Software without prior written authorization from the\n"
             << "Author.\n";
        return 0;
      }
      // '-d' //
      else if (k_same (vsArgs[inArg], "-d"))
      {
        Global::bDebug = true;
      }
      // '-p' //
      else if (k_same (vsArgs[inArg], "-p"))
      {
        Global::bOmitProgress = true;
      }
      // '-s' //
      else if (k_same (vsArgs[inArg], "-s"))
      {
        Global::bPrintLine = true;
      }
      // '-o' //
      else if (k_same (vsArgs[inArg], "-o"))
      {
        if ((inArg+1) < nArgs)
        {
          k_copy (sOutputFilename, vsArgs[++inArg]);
        }
        else
        {
          err (false, EP_MISSING_ARG, "output filename");
          errterm ();
        }
      }
      else
      {
        if (vsArgs[inArg][0] == '-')
        {
          err (false, EP_INVALID_ARG, vsArgs[inArg]);
          errterm ();
        }
        else
        {
          k_copy (sInputFilename, vsArgs[inArg]);
          break;
        }
      }
    }
  }
  else
  {
    cout << "Usage: julc [ option(s) ] [ filename ]\n"
         << "\n"
         << "Options           | Description                               \n"
         << "------------------+-------------------------------------------\n"
         << " -a               | display release and contact info          \n"
         << " -l               | display distribution license              \n"
         << " -p               | do not display progress during compilation\n"
         << " -d               | generate runtime debug information        \n"
         << " -s               | print source code lines in output file    \n"
         << " -o <filename>    | specify output filename                   \n"
         << "\n"
         << "Filename                                                      \n"
         << "--------------------------------------------------------------\n"
         << " * filename of Juliet source file                             \n"
         << " * extension can be omitted (.jul)                            \n";
    return 0;
  }

  // Verify the capture of the filename //
  if (k_same (sInputFilename, "I AM EMPTY!!"))
  {
    err (false, EP_MISSING_ARG, "input filename");
    errterm();
  }

  // Verify the capture of the filename //
  if (k_same (sOutputFilename, "I AM EMPTY!!"))
  {
    err (false, EP_MISSING_ARG, "output filename");
    errterm();
  }

  // Initialize Phase A: Input //
  gInputManager = new InputManager ();
  gInputManager->include (sInputFilename);

  // Initialize Phase B: Lexical Analysis //
  gLexer = new Lexer ();

  // Initialize Phase C: Syntax Analysis //
  gScopeManager = new ScopeManager ();
  gParser = new Parser ();

  // Initialize Phase D: Intermediate Code //
  gICM = new IntermediateCodeManager ();
  gUnique = new UniqueGenerator ();

  if (!Global::bOmitProgress) printf ("Compiling (%s)...\n", sInputFilename);

  // Begin processing //
  gParser->parse (); // parse the program

  if (Global::nErrors <= 0)
  {
    if (!Global::bOmitProgress) printf ("Writing...\n");
    gICM->reset(); // reset the intermediate code position marker
    icoutput (sOutputFilename); // output intermediate code in Romeo Language
  }
 
  if (!Global::bOmitProgress)
  {
    printf ("\n\"%s\" - %d error(s)\n", sOutputFilename, Global::nErrors);
  }

  // Deinitialize Phase D: Intermediate Code //
  delete gICM;
  delete gUnique;

  // Deinitialize Phase C: Syntax Analysis //
  delete gParser;
  delete gScopeManager;

  // Deinitialize Phase B: Lexical Analysis //
  delete gLexer;

  // Deinitialize Phase A: Input //
  delete gInputManager;

  // Deinitialize Global //
  Global::deinitialize();

  k_del_s (sInputFilename);
  k_del_s (sOutputFilename);

  // Leak Tracking //
  #ifdef GLOBAL_DEBUG
    leaktracker_deinit ();
  #endif

  return 0;
}