/***************************************************************************\
  "leaktracker.hpp" - Memory Leak Tracker For Catching Memory Leaks
  Revision 7; 16 Jan 2003 - 15 March 2004
  Copyright (C) 2003-2004, Kevin P. Albrecht
\***************************************************************************/

#ifndef LEAKTRACKER_HPP
#define LEAKTRACKER_HPP

/////////////////////////
// DEFINES //////////////
/////////////////////////

// External Defines //
// GLOBAL_DEBUG or GLOBAL_RELEASE

/////////////////////////
// INCLUDES /////////////
/////////////////////////

// Standard //

// Local //

/////////////////////////
// DECLARATIONS /////////
/////////////////////////

#ifdef GLOBAL_DEBUG

  class LeakTrackerUnit
  {
    public:
      char *file;
      int line;
      void *pointer;
      LeakTrackerUnit* next;
      LeakTrackerUnit* prev;
  };

  LeakTrackerUnit *leaktracker_head;
  unsigned int leaktracker_units;
  unsigned int leaktracker_units_total;

  void leaktracker_init ();
  void leaktracker_deinit ();
  void *leaktracker_alloc (unsigned int size, const char *file, int line);
  void leaktracker_free (void *p);

  inline void* operator new (unsigned int sz, const char *file, int line)
  {
    return leaktracker_alloc (sz, file, line);
  }
  inline void* operator new (unsigned int sz)
  {
    return leaktracker_alloc (sz, "?", 0);
  }

  // This breaks DMC
  #ifndef ENV_DMC
    inline void operator delete (void *p, const char *file, int line)
    {
      leaktracker_free (p);
    }
  #endif

  inline void operator delete (void *p)
  {
    leaktracker_free (p);
  }
  inline void* operator new [] (unsigned int sz, const char *file, int line)
  {
    return leaktracker_alloc (sz, file, line);
  }
  inline void* operator new [] (unsigned int sz)
  {
    return leaktracker_alloc (sz, "?", 0);
  }

  // This breaks DMC
  #ifndef ENV_DMC
    inline void operator delete [] (void *p, const char *file, int line)
    {
      leaktracker_free (p);
    }
  #endif

  inline void operator delete [] (void *p)
  {
    leaktracker_free (p);
  }

  #define new new(__FILE__, __LINE__)

  /////////////////////////
  // DEFINITIONS //////////
  /////////////////////////

  void leaktracker_init ()
  {
    leaktracker_head = NULL;
    leaktracker_units = 0;
    leaktracker_units_total = 0;
  }

  void leaktracker_deinit ()
  {
    LeakTrackerUnit *temp;

    temp = leaktracker_head;

    printf ("**LEAKTRACKER:STATUS** %d memory leak(s) found, %d total units allocatated\n", leaktracker_units, leaktracker_units_total);

    while (leaktracker_head != NULL)
    {
      leaktracker_head = temp->next;

      printf ("**LEAKTRACKER:LEAK** %s(%d), ptr<0x%x>, str<%s>\n", temp->file, temp->line, (int)temp->pointer, (char*)(temp->pointer));

      free (temp->file);
      //NOTE: temp->pointer not deleted here because it should be deleted in the program.
      free (temp);
      temp = leaktracker_head;
    }

    printf ("**LEAKTRACKER:FINISHED**: Anything else after here is fake.\n");
  }

  void *leaktracker_alloc (unsigned int size, const char *file, int line)
  {
    LeakTrackerUnit *temp;
    int n;

    temp = (LeakTrackerUnit *) malloc (sizeof (LeakTrackerUnit));

    n = strlen (file) * sizeof (char);
    temp->file = (char *) malloc (n + 1);
    strcpy (temp->file, file);
    temp->line = line;
    temp->pointer = malloc (size);
    temp->next = leaktracker_head;
    temp->prev = NULL;
    if (leaktracker_head) leaktracker_head->prev = temp;
    leaktracker_head = temp;

    leaktracker_units++;
    leaktracker_units_total++;

    #ifdef GLOBAL_DEBUG_VERBOSE
      printf ("allocating: 0x%x\n", temp->pointer);
    #endif

    return temp->pointer;
  }

  void leaktracker_free (void *p)
  {
    LeakTrackerUnit *curr;

    if (p == NULL) return;

    curr = leaktracker_head;

    while (curr != NULL)
    {
      if (curr->pointer == p)
      {
        free (curr->pointer);
        free (curr->file);

        if (curr->prev == NULL)
        {
          leaktracker_head = curr->next;
          if (leaktracker_head) leaktracker_head->prev = NULL;
        }
        else
        {
          curr->prev->next = curr->next;
          if (curr->next) curr->next->prev = curr->prev;
        }

        free (curr);
        leaktracker_units--;        
        return;
      }

      curr = curr->next;
    }

    printf ("**LEAKTRACKER:CANNOT_FREE**: ptr<0x%x>\n", (int)p);
  }

#endif /* GLOBAL_DEBUG */

#endif
